###################
# load R packages
###################

install.packages("VAR.etp")
require(VAR.etp)

###################################
# modify the VAR.BPR function
####################################

VAR.BPR = function (x, p, h, nboot = 500, type = "const", alpha = 0.95) 
{ 
  resamp<- VAR.etp:::resamp
  VARB.Boot<- VAR.etp:::VARB.Boot
  VAR.est<- VAR.etp:::VAR.est
  VAR.adjust<- VAR.etp:::VAR.adjust
  VARB.ys<- VAR.etp:::VARB.ys
  VAR.ForeB<- VAR.etp:::VAR.ForeB
  VAR.PI<- VAR.etp:::VAR.PI
  VARB.est<-VAR.etp:::VARB.est
  k <- ncol(x)
  var1 <- VAR.est(x, p, type)
  var2 <- VARB.est(x, p, type)
  b1 <- var1$coef
  e1 <- var1$resid
  b2 <- var2$coef
  e2 <- var2$resid
  f <- VAR.Fore(x, b1, p, h, type = type)
  fstat <- array(0, c(h, k, nboot))
  for(i in 1:nboot) 
  {
    es <- resamp(e2)
    ys <- VARB.ys(x, b2, p, es, type)
    vars <- VAR.est(ys, p)
    bs <- vars$coef
    fs <- VAR.ForeB(x, bs, p, h, e1)
    fstat[, , i] <- as.matrix(fs)
  }
  PIs <- VAR.PI(fstat, alpha, f)$PI
  return(list(Intervals = PIs, boot.sample=fstat,Forecast = f, Prob = alpha))
}

####################################
# FIT FVar ON DEMOGAPHY DATA
####################################

fvar <- function(demogdata, k)
{
  epsilon.f <- demogdata$smooth.err$female
  epsilon.m <- demogdata$smooth.err$male
  epsilon.f[is.na(epsilon.f)] <- 0
  epsilon.m[is.na(epsilon.m)] <- 0
  smooth.var <- list(female = epsilon.f, male = epsilon.m)
  y.f <- log(demogdata$rate$female)
  y.m <- log(demogdata$rate$male)
  fts.obj.f <- fts(demogdata$age, y.f)
  fts.obj.m <- fts(demogdata$age, y.m)
  pca.f <- ftsm(fts.obj.f, order = k)
  pca.m <- ftsm(fts.obj.m, order = k)
  
  var.f <- vars::VAR(pca.f$coeff[, 2:(k+1)])
  var.m <- vars::VAR(pca.m$coeff[, 2:(k+1)])
  
  # fitted curves
  fitted.curve.f <- pca.f$basis[,2:(k+1)]%*%t(fitted(var.f)) + pca.f$basis[, 1]
  fitted.curve.m <- pca.m$basis[,2:(k+1)]%*%t(fitted(var.m)) + pca.m$basis[, 1]
  fitted <- list(female = fitted.curve.f, male = fitted.curve.m)
  
  # residual curves
  resid.f <- fitted.curve.f - y.f[, -1]
  resid.m <- fitted.curve.m - y.m[, -1]
  residuals <- list(female = resid.f, male = resid.m)
  basis <- list(female = pca.f$basis, male = pca.m$basis)
  coeffs <- list(female = pca.f$coeff, male = pca.m$coeff)
  return(list(dim = k, fitted = fitted, residuals = residuals,
              age = demogdata$age, year = demogdata$year, smooth.var = smooth.var, 
              coeffs = coeffs, basis = basis))
}

####################################
# FORECAST FVAR
####################################

forecast.fvar <- function(object, h = 20, level = 0.8, nboot = 100)
{
  n <- length(object$year)
  m <- length(object$age)
  k <- object$dim
  boot.curve.f <- boot.curve.m <- array(NA, dim = c(m, h, nboot))
  fore.obj.f <- fore.obj.m <- array(NA, dim = c(age+1, h))
  
  # VAR (constant)
  fore.f <- VAR.BPR(object$coeffs$female[,2:(k+1)], 2, h = h, nboot = nboot, type = "const",
                   alpha = level)
  fore.m <- VAR.BPR(object$coeffs$male[,2:(k+1)], 2, h = h, nboot = nboot,
                   type = "const", alpha = level)
  fore.obj.f <- object$basis$female[,2:(k+1)] %*% t(fore.f$Forecast) + object$basis$female[,1] 
  fore.obj.m <- object$basis$male[,2:(k+1)] %*% t(fore.m$Forecast) + object$basis$male[,1] 
  
  # bootstrapped curves
  for(b in 1:nboot)
  {
    ind.f = sample(1:(38+iw), h, replace = TRUE)
    ind.m = sample(1:(38+iw), h, replace = TRUE)
    boot.curve.f[,,b] = object$basis$female[,2:(k+1)] %*% t(fore.f$boot.sample[,,b]) + 
                        object$basis$female[,1] + object$residuals$female[, ind.f]+ matrix(rnorm(m*h, 0, rep(object$smooth.var$female[,ncol(object$smooth.var$female)], h)),m, h)
    boot.curve.m[,,b] = object$basis$male[,2:(k+1)] %*% t(fore.m$boot.sample[,,b]) + 
                        object$basis$male[,1] + object$residuals$male[, ind.m]+ matrix(rnorm(m*h, 0, rep(object$smooth.var$male[,ncol(object$smooth.var$male)], h)),m, h)
  }
  boot.sample <- list(female = boot.curve.f, male = boot.curve.m)
  Forecast <- list(female = fore.obj.f, male = fore.obj.m)
  pi.f <- VAR.etp:::VAR.PI(boot.curve.f, level, fore.obj.f)$PI
  pi.m <- VAR.etp:::VAR.PI(boot.curve.m, level, fore.obj.m)$PI
  PIs <- list(female = pi.f, male = pi.m)
  return(list(boot.sample = boot.sample, Forecast = Forecast, Intervals = PIs))
}

